import z3

import DPGM as AF
import z3_instance as zi
import labeling_scheme as ls


class Labeling:
    """
    A mapping of the arguments from an AF to one of the labels 'in, 'out' or 'undec'.
    String representation is the sets of arguments labeled 'in', 'out' and 'undec' in that order, i.e. '{A,B}{C}{}' for
    A and B labeled 'in', C labeled 'out' and no argument labeled 'undec'.
    """

    def __init__(self, labeling_scheme: ls.LabelingScheme, z3i: zi.Z3Instance, model: z3.ModelRef):
        self.af = z3i.af
        self.z3i = z3i
        self.model = model
        self.labeling_scheme = labeling_scheme

        self.label_dict = dict()
        self.labeled_in = []
        self.labeled_out = []
        self.labeled_undec = []

        self._compute_labeling()

    def _compute_labeling(self):
        for argument in self.af.get_nodes():
            label = self.labeling_scheme.get_label(argument, self.model, self.z3i)
            self.label_dict[argument.id] = label
            if label == ls.l_in:
                self.labeled_in.append(argument)
            elif label == ls.l_out:
                self.labeled_out.append(argument)
            elif label == ls.l_undec:
                self.labeled_undec.append(argument)
            else:
                assert False, "unknown label " + label

    def get_label(self, argument: AF.Node) -> str:
        return self.label_dict[argument.id]

    def get_in(self):
        """
        Returns the list of all nodes labelled 'in'.
        """
        return self.labeled_in

    def get_out(self):
        """
        Returns the list of all nodes labelled 'out'.
        """
        return self.labeled_out

    def get_undec(self):
        """
        Returns a list of all nodes labelled 'undec'.
        """
        return self.labeled_undec

    def get_labeling_triple(self):
        """
        Returns the lists of arguments labelled 'in', 'out' and 'undec' in that order as tuple.
        """
        return self.labeled_in, self.labeled_out, self.labeled_undec

    def __repr__(self):
        repr_set = lambda s: "{" + ",".join(sorted(map(repr, s))) + "}"
        set_in = repr_set(self.labeled_in)
        set_out = repr_set(self.labeled_out)
        set_undec = repr_set(self.labeled_undec)
        return set_in + set_out + set_undec

    def __eq__(self, other):
        return isinstance(other, Labeling) and self.get_labeling_triple() == other.get_labeling_triple()

    def __hash__(self):
        return hash(repr(self))

    def __lt__(self, other):
        return self.get_labeling_triple() < other.get_labeling_triple()



from typing import List, Set

import DPGM as AF


class ClassicalSemantics:

    def __init__(self, af: AF.DPGM):
        self.af = af

    def get_extensions(self) -> List[Set[AF.Node]]:
        # implemented by subclasses
        pass


class CF(ClassicalSemantics):
    """Conflict-freeness: A set of arguments S is conflict free iff there is no A and B in S with A->B or B->A."""
    def get_extensions(self):
        extension_candidates = [set()]
        for node in self.af.get_nodes():
            conflicting_nodes = node.get_parents().union(node.get_children())
            new_extension_candidates = []
            for candidate in extension_candidates:
                if not conflicting_nodes.intersection(candidate):
                    new_candidate = candidate.copy()
                    new_candidate.add(node)
                    new_extension_candidates.append(new_candidate)
            extension_candidates.extend(new_extension_candidates)
        return extension_candidates


def defends(af: AF.DPGM, nodes: Set[AF.Node]):
    """
    A node is defended by a set of arguments if all its attackers are in turn attacked from within the set.
    :return: the set of nodes defended by the given set
    """
    defended = set()
    for node in af.get_nodes():
        if node.is_initial():
            defended.add(node)
            continue
        # check whether all attackers of the node are attacked from within the given set of nodes
        all_attacked = True
        for attacker in node.get_parents():
            if not attacker.get_parents().intersection(nodes):
                all_attacked = False
                break
        if all_attacked:
            defended.add(node)
    return defended


class Admissible(ClassicalSemantics):
    """
    A set of arguments is admissible if it is conflict free and defends itself.
    """
    def get_extensions(self):
        scf = CF(self.af)
        adm_extensions = []
        for candidate in scf.get_extensions():
            if candidate.issubset(defends(self.af, candidate)):
                adm_extensions.append(candidate)
        return adm_extensions


class Complete(ClassicalSemantics):
    """
    A set of arguments is complete if it is conflict free and defends exactly itself, i.e. the set equals the set of
    arguments it defends.
    """
    def get_extensions(self):
        scf = CF(self.af)
        cmp_extensions = []
        for candidate in scf.get_extensions():
            if candidate == defends(self.af, candidate):
                cmp_extensions.append(candidate)
        return cmp_extensions


class Grounded(ClassicalSemantics):
    """
    The grounded extension is the complete extension which is minimal w.r.t. set inclusion among all complete
    extensions. It always exists and is unique.
    """
    def get_extensions(self):
        cmp = Complete(self.af)
        cmp_extensions = cmp.get_extensions()
        grn_extensions = []
        for candidate in cmp_extensions:
            if all(map(candidate.issubset, cmp_extensions)):
                grn_extensions.append(candidate)
        assert len(grn_extensions) == 1
        return grn_extensions


class Preferred(ClassicalSemantics):
    """
    A set of nodes is a preferred extension if it is a complete extension and maximal w.r.t. set inclusion among all
    complete extensions. There can be multiple preferred extensions.
    """
    def get_extensions(self):
        cmp = Complete(self.af)
        cmp_extensions = cmp.get_extensions()
        prf_extensions = []
        for candidate in cmp_extensions:
            if all(map(lambda s: not candidate.issubset(s) or candidate == s, cmp_extensions)):
                prf_extensions.append(candidate)
        return prf_extensions


class SemiStable(ClassicalSemantics):
    def get_extensions(self):
        raise NotImplemented


class Stable(ClassicalSemantics):
    def get_extensions(self):
        raise NotImplemented



from typing import List, Set

import z3

import DPGM as AF
from labeling import Labeling
from labeling_scheme import LabelingScheme
from semantics import Semantics
from z3_instance import Z3Instance


def get_model(z3i: Z3Instance, semantics: List[Semantics], constraints=None):
    """
    Check if a distribution satisfying the constraints of all semantics plus potential additional constraints exists.
    If so, return it as a model.

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param constraints: optional list of additional constraints
    :return: A model if one exists, None otherwise
    """
    if not constraints:
        constraints = []

    for sem in semantics:
        # print(sem.__class__.__name__, sem.get_constraints())
        constraints.extend(sem.get_constraints())

    result = z3i.run(constraints)
    if result == z3.sat:
        return z3i.solver.model()
    else:
        return None


def get_satisfying_labeling(z3i: Z3Instance, lab: LabelingScheme, semantics: List[Semantics]) -> Labeling:
    """
    Get a labeling satisfying the given semantics under the given labeling scheme. Returns None if no such labeling
    exists.

    :param z3i: the underlying z3 instance with the AF
    :param lab: a labeling scheme
    :param semantics: a list of semantics
    :return: A labeling or None
    """
    model = get_model(z3i, semantics)
    if model:
        return Labeling(lab, z3i, model)


def get_all_satisfying_labelings(z3i: Z3Instance, lab: LabelingScheme, semantics: List[Semantics]) -> Set[Labeling]:
    """
    Compute all labelings under the given labeling scheme that satisfy the given semantics.

    :param z3i: the underlying z3 instance with the AF
    :param lab: a labeling scheme
    :param semantics: a list of semantics
    :return: a list of labelings
    """
    for sem in semantics:
        # print(sem.__class__.__name__, sem.get_constraints())
        z3i.add_constraints(sem.get_constraints())
    # print("z3i.constraints: ", z3i.constraints)
    # test overall satisfiability
    result = z3i.run([])
    print("Overall satisfiability:", result)
    if result != z3.sat:
        return set()

    candidate_constraints = [[]]
    labelings = set()
    num_of_checks = 1

    # num_nodes = len(z3i.af.get_nodes())
    for node in z3i.af.get_nodes():
        print(node.name)
        next_candidate_constraints = []
        for candidate_constraint in candidate_constraints:
            for const in lab.get_constraints(node, z3i):
                new_candidate_constraint = candidate_constraint.copy()
                new_candidate_constraint.append(const)
                # add candidate constraints to solver
                z3i.solver.push()  # this adds an backtracking point
                # print(z3i.solver)
                z3i.solver.add(new_candidate_constraint)
                # print(z3i.solver)
                result = z3i.solver.check()
                num_of_checks += 1
                if result == z3.sat:
                    # print("new_candidate_constraint", new_candidate_constraint)
                    next_candidate_constraints.append(new_candidate_constraint)
                    model = z3i.solver.model()
                    labeling = Labeling(lab, z3i, model)
                    labelings.add(labeling)
                    # print("model", model)
                    # print("labeling:", labeling)
                #  remove candidate constraints from solver
                z3i.solver.pop()  # go back to the backtracking point
        candidate_constraints = next_candidate_constraints

    print("Finished. Performed", num_of_checks, "checks.")
    return labelings


def check_credulous_threshold_acceptance(z3i: Z3Instance, semantics: List[Semantics], threshold: float,
                                         argument: AF.Node):
    """
    Under a given threshold, check if the given argument is credulously accepted, i.e. there exists at least one
    distribution satisfying all given semantics that assigns the argument a probability greater or equal to the
    threshold.

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param threshold: a float from the interval [0,1]
    :param argument: an argument from the AF
    :return: True + a model if one exists, False otherwise
    """
    argument_prob = z3i.get_node_prob_var(argument)
    constraints = [argument_prob >= threshold]
    model = get_model(z3i, semantics, constraints)
    if model:
        return True, model
    else:
        return False, None


def check_credulous_acceptance(z3i: Z3Instance, semantics: List[Semantics], argument: AF.Node):
    """
    Check if the given argument is credulously accepted, i.e. there exists at least one distribution satisfying all
    given semantics that assigns the argument probability one.

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param argument: an argument from the AF
    :return: True + a model if one exists, False otherwise
    """
    threshold = 1
    return check_credulous_threshold_acceptance(z3i, semantics, threshold, argument)


def check_skeptical_threshold_acceptance(z3i: Z3Instance, semantics: List[Semantics], threshold: float,
                                         argument: AF.Node):
    """
    Under a given threshold, check if the given argument is skeptically accepted, i.e. all distributions that satisfy
    all given semantics assign the argument a probability greater or equal to the threshold.

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param threshold: a float from the interval [0,1]
    :param argument: an argument from the AF
    :return: True or False + counter example
    """
    # "check if for all satisfying distributions, P(A) >= t holds"
    # <=> not "check if for one satisfying distribution, P(A) < t holds"
    argument_prob = z3i.get_node_prob_var(argument)
    constraints = [argument_prob < threshold]
    model = get_model(z3i, semantics, constraints)
    if model:
        return False, model
    else:
        return True, None


def check_skeptical_acceptance(z3i: Z3Instance, semantics: List[Semantics], argument: AF.Node):
    """
    Check if the given argument is skeptically accepted, i.e. all distributions that satisfy all given semantics assign
    the argument probability one.

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param argument: an argument from the AF
    :return: True or False + counter example
    """
    threshold = 1
    return check_skeptical_threshold_acceptance(z3i, semantics, threshold, argument)



import argparse

import af_parser
import labeling_scheme as lab
import semantics as sem
import tasks
import z3_instance
from timer import Timer

# TODO:
#  - Flag for extensions
#  - flag to generate unsat core; this needs named assertions, see assert_and_track() in z3.py
#  - check why seemingly the same labeling still can appear multiple times for AL
#  - af_constraints in z3i overlaps with AF semantics
#  - credulous and skeptical acceptance wrt. a labeling scheme


def main():
    main_timer = Timer(output_string="Overall runtime:")

    #########################
    # set up argument parser
    #########################
    arg_parser = argparse.ArgumentParser(description="CPrAA - a Checker for Probabilistic Abstract Argumentation ",
                                         add_help=False)
    base_group = arg_parser.add_argument_group("basic parameters")
    base_group.add_argument("-f", "--file", help="the argumentation framework as .tgf file")
    base_group.add_argument("-s", "--semantics", nargs="+", default=[],
                            help="the probabilistic semantics to check against")
    base_group.add_argument("-OD", "--one_distribution", action="store_true",
                            help="compute one distribution satisfying the semantics")
    base_group.add_argument("-i", "--assume_independence", action="store_true",
                            help="treat the argument's probabilities as independent (warning: this leads to incorrect "
                                 "results for some semantics)")

    ac_group = arg_parser.add_argument_group("acceptance checking")
    ac_group.add_argument("-a", "--argument", help="the argument from the AF to be checked")
    ac_group.add_argument("-CA", "--credulous_acceptance",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check credulous acceptance of given argument, "
                               "optionally taking a threshold into account")
    ac_group.add_argument("-SA", "--skeptical_acceptance",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check skeptical acceptance of given argument, "
                               "optionally taking a threshold into account")
    ac_group.add_argument("-CAA", "--credulous_acceptance_all",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check credulous acceptance of all arguments, "
                               "optionally taking a threshold into account")
    ac_group.add_argument("-SAA", "--skeptical_acceptance_all",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check skeptical acceptance of all arguments, "
                               "optionally taking a threshold into account")

    lab_group = arg_parser.add_argument_group("generate labelings")
    lab_group.add_argument("-l", "--labeling_scheme", metavar="SCHEME", help="the labeling scheme to use")
    lab_group.add_argument("-OL", "--one_labeling", action="store_true",
                           help="compute one labeling satisfying the semantics")
    lab_group.add_argument("-AL", "--all_labelings", action="store_true",
                           help="compute all labelings satisfying the semantics")
    lab_group.add_argument("-lt", "--labeling_threshold", nargs="+", metavar="T", type=float,
                           help="the optional thresholds needed for some labeling schemes")

    info_group = arg_parser.add_argument_group('informational parameters')
    info_group.add_argument("-h", "--help", action="help", help="show this help message")
    info_group.add_argument("-ls", "--list_semantics", action="store_true",
                            help="list all available probabilistic semantics")
    info_group.add_argument("-ll", "--list_labeling_schemes", action="store_true",
                            help="list all available labeling schemes")
    info_group.add_argument("-d", "--documentation", nargs="+", default=[], metavar="NAME",
                            help="display a short documentation for the listed semantics or labeling schemes")
    info_group.add_argument("-t", "--time", action="store_true", help="time the execution")
    info_group.add_argument("-o", "--distribution_output_format", metavar="FORMAT", type=str, default="M",
                            help="how to output distributions: (M) marginal node probabilities, (F) full distribution, "
                                 "(S) support of distribution, (Z) Z3 output, (T) SMT2 format.")

    ##################
    # parse arguments
    ##################
    args = arg_parser.parse_args()

    af = None
    z3i = None
    labeling_scheme = None
    semantics = []
    argument = None

    main_timer.enable_printing = args.time
    Timer.enable_printing_default = args.time

    if args.list_semantics:
        print("Available semantics:", ", ".join(sem.all_semantics_names()))

    if args.list_labeling_schemes:
        print("Available labeling schemes:", ", ".join(lab.get_all_labeling_scheme_names()))

    if args.documentation:
        for name in args.documentation:
            semantics_class = sem.get_semantics_class_by_name(name)
            if semantics_class:
                print("Semantics '" + semantics_class.__name__ + "':")
                if semantics_class.__doc__:
                    print(semantics_class.__doc__.strip("\n"))
                else:
                    print("    No documentation available.")
            else:
                labeling_scheme_class = lab.get_labeling_scheme_class(name)
                if labeling_scheme_class:
                    print("Labeling scheme '" + labeling_scheme_class.__name__ + "':")
                    if labeling_scheme_class.__doc__:
                        print(labeling_scheme_class.__doc__.strip("\n"))
                    else:
                        print("    No documentation available.")
                else:
                    arg_parser.error("no semantics or labeling scheme called '" + name + "' exists")

    if args.file:
        af = af_parser.parse_tgf(args.file)
        af.print()
        z3i_setup_timer = Timer(output_string="z3i setup took")
        z3i = z3_instance.Z3Instance(af, args.assume_independence)
        z3i_setup_timer.stop()

    if args.semantics:
        if not af:
            arg_parser.error("an AF must be provided when checking semantics")

        semantics_constraints_timer = Timer(output_string="Generating constraints for all semantics took")
        for semantics_name in args.semantics:
            semantics_class = sem.get_semantics_class_by_name(semantics_name)
            if not semantics_class:
                arg_parser.error("No semantics called '"
                                 + semantics_name + "' exists. Use -ls to list all available semantics.")
            semantics.append(semantics_class(z3i))

        semantics_constraints_timer.stop()

    if args.labeling_scheme:
        labeling_scheme_class = lab.get_labeling_scheme_class(args.labeling_scheme)
        if not labeling_scheme_class:
            arg_parser.error("no labeling scheme called '" + args.labeling_scheme + "' exists")
        num_args_required = labeling_scheme_class.num_args
        if args.labeling_threshold:
            num_args_given = len(args.labeling_threshold)
        else:
            num_args_given = 0
        if num_args_given < num_args_required:
            error_message = "mismatch in the number of thresholds: required by labeling scheme '" \
                            + args.labeling_scheme + "': " + str(num_args_required) + ", given: " + str(num_args_given)
            arg_parser.error(error_message)
        if num_args_given > num_args_required:
            warning_message = "warning: mismatch in the number of thresholds: required by labeling scheme '" \
                              + args.labeling_scheme + "': " + str(num_args_required) + ", given: " \
                              + str(num_args_given) + " - i'm ignoring the surplus ones"
            print(warning_message)
        if num_args_required == 0:
            labeling_scheme = labeling_scheme_class()
        elif num_args_required == 1:
            labeling_scheme = labeling_scheme_class(args.labeling_threshold[0])
        elif num_args_required == 2:
            labeling_scheme = labeling_scheme_class(args.labeling_threshold[0], args.labeling_threshold[1])
        else:
            raise NotImplementedError("labeling schemes with more than 2 thresholds")

    if args.argument:
        if not af:
            arg_parser.error("an AF must be provided when checking acceptance")
        argument = af.get_node_by_name(args.argument)

    ########
    # tasks
    ########

    if args.one_distribution:
        if not af:
            arg_parser.error("an AF must be provided to compute a distribution")

        od_timer = Timer(output_string="Computing one distribution took")
        print("Computing one distribution satisfying the following semantics:",
              ", ".join(args.semantics))
        model = tasks.get_model(z3i, semantics)
        if model:
            z3i.print_formatted_distribution(model, args.distribution_output_format)
        else:
            print("No distribution satisfying the given semantics exists.")
        od_timer.stop()

    if args.one_labeling:
        if not af:
            arg_parser.error("an AF must be provided to compute a labeling")
        if not labeling_scheme:
            arg_parser.error("no labeling provided")

        ol_timer = Timer(output_string="Computing one labeling took")
        print("Computing one", args.labeling_scheme, "labeling satisfying the following semantics:",
              ", ".join(args.semantics))
        labeling = tasks.get_satisfying_labeling(z3i, labeling_scheme, semantics)
        print(labeling)
        ol_timer.stop()

    if args.all_labelings:
        if not af:
            arg_parser.error("an AF must be provided to compute a labeling")
        if not labeling_scheme:
            arg_parser.error("no labeling provided")

        al_timer = Timer(output_string="Computing all labelings took")
        print("Computing all", args.labeling_scheme, "labelings satisfying the following semantics:",
              ", ".join(args.semantics))
        labelings = tasks.get_all_satisfying_labelings(z3i, labeling_scheme, semantics)
        print("Number of labelings:", len(labelings))
        # for ext in sorted([labeling.labeled_in for labeling in labelings]):
        #     print(ext)
        # for labeling in labelings:
        #     print(labeling)
        #     # print(labeling.model)
        #     # z3i.print_distribution(labeling.model, only_positive=True)
        #     # z3_instance.print_model(af, labeling.model)
        for labeling in sorted(list(set(map(repr, labelings)))):
            print(labeling)
        al_timer.stop()

    if args.credulous_acceptance is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")
        if not argument:
            arg_parser.error("argument to be checked is missing")

        threshold = args.credulous_acceptance

        ca_timer = Timer(output_string="Checking credulous acceptance took")
        print("Checking credulous acceptance of argument", args.argument, "with threshold", threshold,
              "under the following semantics:", ", ".join(args.semantics))
        (acceptable, model) = tasks.check_credulous_threshold_acceptance(z3i, semantics, threshold, argument)
        if acceptable:
            print(args.argument, "is credulously accepted.")
            print("Satisfying distribution:")
            z3i.print_formatted_distribution(model, args.distribution_output_format)
            # z3i.print_distribution(model, only_positive=True)
            # z3_instance.print_model(af, model)
        else:
            print(args.argument, "is not credulously accepted.")
        ca_timer.stop()

    if args.skeptical_acceptance is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")
        if not argument:
            arg_parser.error("argument to be checked is missing")

        threshold = args.skeptical_acceptance

        sa_timer = Timer(output_string="Checking skeptical acceptance took")
        print("Checking skeptical acceptance of argument", args.argument, "with threshold", threshold,
              "under the following semantics:", ", ".join(args.semantics))
        (acceptable, counterexample) = tasks.check_skeptical_threshold_acceptance(z3i, semantics, threshold, argument)
        if acceptable:
            print(args.argument, "is skeptically accepted.")
        else:
            print(args.argument, "is not skeptically accepted.")
            print("Counterexample:")
            z3i.print_formatted_distribution(counterexample, args.distribution_output_format)
            # z3i.print_distribution(counterexample, only_positive=True)
            # z3_instance.print_model(af, counterexample)
        sa_timer.stop()

    if args.credulous_acceptance_all is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")

        threshold = args.credulous_acceptance_all

        ca_timer = Timer(output_string="Checking credulous acceptance for all arguments took")
        print("Checking credulous acceptance of all arguments with threshold", threshold,
              "under the following semantics:", ", ".join(args.semantics))
        accepted_args = []
        not_accepted_args = []
        for argument in af.get_nodes():
            (acceptable, model) = tasks.check_credulous_threshold_acceptance(z3i, semantics, threshold, argument)
            if acceptable:
                accepted_args.append(argument)
                # print(argument.name, "is credulously accepted.")
                # print("Satisfying distribution:")
                # z3i.print_distribution(model, only_positive=True)
                # z3_instance.print_model(af, model)
            else:
                not_accepted_args.append(argument)
                # print(argument.name, "is not credulously accepted.")
        print("Credulously accepted:", accepted_args)
        print("Not credulously accepted:", not_accepted_args)
        ca_timer.stop()

    if args.skeptical_acceptance_all is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")

        threshold = args.skeptical_acceptance_all

        ca_timer = Timer(output_string="Checking skeptical acceptance for all arguments took")
        print("Checking skeptical acceptance of all arguments with threshold", threshold,
              "under the following semantics:", ", ".join(args.semantics))
        accepted_args = []
        not_accepted_args = []
        for argument in af.get_nodes():
            acceptable, counterexample = tasks.check_skeptical_threshold_acceptance(z3i, semantics, threshold, argument)
            if acceptable:
                accepted_args.append(argument)
                # print(argument.name, "is skeptically accepted.")
            else:
                not_accepted_args.append(argument)
                # print(argument.name, "is not skeptically accepted.")
                # print("Counterexample:", counterexample)
                # z3i.print_distribution(counterexample, only_positive=True)
                # z3_instance.print_model(af, counterexample)
        print("Skeptically accepted:", accepted_args)
        print("Not skeptically accepted:", not_accepted_args)
        ca_timer.stop()

    main_timer.stop()

    # CSV friendly print of timing results
    # semantics_names = ",".join([semantics_obj.__class__.__name__ for semantics_obj in semantics])
    # timings = ";".join([str(timer.last_timing) for timer in Timer.all_timers])
    # # print(semantics_names + ";" + timings)
    # print(af.name + ";" + timings)


main()


# ---------------------------------------------------------

def test_all_semantics():
    # path = "AFs/af01.tgf"
    path = "AFs/af_references_noisy.tgf"
    af = af_parser.parse_tgf(path)
    z3i = z3_instance.Z3Instance(af)

    for semantics_class in tasks.Semantics.__subclasses__():
        try:
            semantics = semantics_class(z3i)
            print(semantics_class.__name__, semantics.constraints)
        except NotImplementedError:
            print(semantics_class.__name__, "(not implemented)")


# test_all_semantics()



from typing import List, Tuple, Any

Assignment = List[Tuple[Any, Any]]  # type alias

DEFAULT_PRECISION = 5


def generate(ar: list, br: list = (False, True)) -> List[Assignment]:
    """
    Generates all possible assignments of the elements in list 'ar' to elements of list 'br'.
    E.g. ar = [x,y], br = [0,1] -> [[(x,0),(y,0)], [(x,0),(y,1)], [(x,1),(y,0)], [(x,1),(y,1)]]
    """
    combinations = [[]]

    for a in ar:
        next_combinations = []
        for c in combinations:
            for b in br:
                new_c = c.copy()
                new_c.append((a, b))
                next_combinations.append(new_c)
        combinations = next_combinations
    return combinations


def repr_assignment(a: Assignment) -> str:
    strings = []
    for (node, b) in a:
        strings.append((" " if b else "-") + str(node))
    return ",".join(strings)


def repr_assignment_prob(a: Assignment, prob: float, given: Assignment = None, precision=DEFAULT_PRECISION):
    cond = "P(" + repr_assignment(a) + ("|" + repr_assignment(given) if given else "") + ")"
    return cond + " = " + "{:0.{p}g}".format(prob, p=precision)


def print_assignment_prob(a: Assignment, prob: float, precision=DEFAULT_PRECISION):
    print(repr_assignment_prob(a, prob, [], precision))



from typing import List, Type

import z3

import assignments
import classical_semantics as classical
import DPGM as AF
import z3_instance as zi


class Semantics:
    def __init__(self, z3_instance: zi.Z3Instance):
        self.context = z3_instance.context
        self.af = z3_instance.af
        self.z3i = z3_instance
        self.constraints: List[z3.BoolRef] = []

        self.generate_constraints()

    def generate_constraints(self):
        # implemented by subclasses
        pass

    def get_constraints(self):
        return self.constraints


####################
# trivial semantics
####################

class Min(Semantics):
    """
    Minimality semantics: All nodes must hold with probability 0.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            self.constraints.append(node_prob == 0)


class Neu(Semantics):
    """
    Neutrality semantics: All nodes must hold with probability 0.5.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            self.constraints.append(node_prob == 0.5)


class Max(Semantics):
    """
    Maximality semantics: All nodes must hold with probability 1.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            self.constraints.append(node_prob == 1)


class Dirac(Semantics):
    """
    Dirac semantics: All nodes must hold with either probability 0 or 1.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            self.constraints.append(z3.Or(node_prob == 0, node_prob == 1))


class Ter(Semantics):
    """
    Ternary semantics: All nodes must hold with either probability 0, 0.5 or 1.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            self.constraints.append(z3.Or(node_prob == 0, node_prob == 0.5, node_prob == 1))


################################
# semantics by Hunter and Thimm
################################

class Fou(Semantics):
    """
    Foundedness semantics: Initial nodes must hold with probability 1.
    """
    def generate_constraints(self):
        for node in self.af.get_initial_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            self.constraints.append(node_prob == 1)


class SFou(Semantics):
    """
    Semi-Foundedness semantics: Initial nodes must hold with probability >= 0.5.
    """
    def generate_constraints(self):
        for node in self.af.get_initial_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            self.constraints.append(node_prob >= 0.5)


class Inv(Semantics):
    """
    Involution semantics: p_A = 1 - p_B for all attacks A -> B.
    """
    def generate_constraints(self):
        for edge in self.af.get_edges():
            prob_node_from = self.z3i.get_node_prob_var(edge.node_from)
            prob_node_to = self.z3i.get_node_prob_var(edge.node_to)
            self.constraints.append(prob_node_from == 1 - prob_node_to)


class Rat(Semantics):
    """
    Rationality semantics: p_A > 0.5 implies p_B <= 0.5 for all attacks A -> B.
    """
    def generate_constraints(self):
        for edge in self.af.get_edges():
            prob_node_from = self.z3i.get_node_prob_var(edge.node_from)
            prob_node_to = self.z3i.get_node_prob_var(edge.node_to)
            self.constraints.append(z3.Implies(prob_node_from > 0.5, prob_node_to <= 0.5, ctx=self.context))


class Coh(Semantics):
    """
    Coherency semantics: p_A <= 1 - p_B for all attacks A -> B.
    """
    def generate_constraints(self):
        for edge in self.af.get_edges():
            prob_node_from = self.z3i.get_node_prob_var(edge.node_from)
            prob_node_to = self.z3i.get_node_prob_var(edge.node_to)
            self.constraints.append(prob_node_from <= 1 - prob_node_to)


class Opt(Semantics):
    """
    Optimism semantics: p_A >= 1 - Sum p_B for all attackers B of A.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.is_initial():
                self.constraints.append(node_prob == 1)
                continue
            summation = z3.Sum([self.z3i.get_node_prob_var(attacker) for attacker in node.get_parents()])
            self.constraints.append(node_prob >= 1 - summation)


class SOpt(Semantics):
    """
    Semi-optimism semantics: p_A >= 1 - Sum p_B for all attackers B of A iff A has at least one attacker.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.is_initial():
                continue
            node_prob = self.z3i.get_node_prob_var(node)
            summation = z3.Sum([self.z3i.get_node_prob_var(attacker) for attacker in node.get_parents()])
            self.constraints.append(node_prob >= 1 - summation)


class Jus(Semantics):
    """
    Justifiability semantics: Coherency and optimism constraints hold.
    """
    def generate_constraints(self):
        opt = Opt(self.z3i)
        coh = Coh(self.z3i)
        self.constraints.extend(opt.get_constraints() + coh.get_constraints())


######################
# semantics by Baier
######################

class CF(Semantics):
    """
    Conflict-freeness semantics: p_A_B = 0 for all attacks A -> B.
    """
    def generate_constraints(self):
        for edge in self.af.get_edges():
            assignment = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = self.z3i.get_prob_var(assignment)
            self.constraints.append(prob_var == 0)


# Admissibility

def almost_sure_defense_constraint(z3i: zi.Z3Instance, node: AF.Node):
    """
    Generate a constraint for an argument A to be almost-surely defended.
    This means for all attackers B->A, the probability that B is in turn attacked (and thus A is defended) is 1.
    That is, P(C1 or C2 or ... or Cn) = 1 for attackers C1...Cn of B, or, equivalently, P(nC1, nC2, ..., nCn) = 0.
    If an attacker B of A has no attackers, then A cannot be almost-surely defended, thus the impossible constraint
    'False' is added.
    """
    parents = list(node.get_parents())
    if len(parents) == 0:
        # unattacked nodes are necessarily defended
        return z3.BoolVal(True, ctx=z3i.context)

    # special treatment for the first attacker as starting point for nested Ands
    first_attacker = parents[0]
    first_attacker_parents = first_attacker.get_parents()
    if first_attacker_parents:
        assignment = []
        for defender in first_attacker_parents:
            assignment.append((defender, False))
        prob_var = z3i.get_prob_var(assignment)
        constraint = prob_var == 0
    else:
        return z3.BoolVal(False, ctx=z3i.context)

    for attacker in parents[1:]:
        attacker_parents = attacker.get_parents()
        if attacker_parents:
            assignment = []
            for defender in attacker_parents:
                assignment.append((defender, False))
            prob_var = z3i.get_prob_var(assignment)
            constraint = z3.And(constraint, prob_var == 0)
        else:
            return z3.BoolVal(False, ctx=z3i.context)
    return constraint


class WAdm(Semantics):
    """
    Weak admissibility semantics: CF and p_A = 1 implies A is almost-surely defended.
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            if node.is_initial():
                # initial nodes are always almost-surely defended, thus the constraint would become 'p_A == 1 -> True'
                continue
            node_prob = self.z3i.get_node_prob_var(node)
            as_defense_constraint = almost_sure_defense_constraint(self.z3i, node)
            self.constraints.append(z3.Implies(node_prob == 1, as_defense_constraint, ctx=self.context))


class LabAdm(Semantics):
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            or_list = []
            for parent_node in node.get_parents():
                parent_node_prob = self.z3i.get_node_prob_var(parent_node)
                or_list.append(parent_node_prob == 1)
            self.constraints.append(z3.Implies(node_prob == 0, z3.Or(or_list, self.context), self.context))


class SLabAdm(Semantics):
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.is_initial():
                continue
            or_list = []
            for parent_node in node.get_parents():
                parent_node_prob = self.z3i.get_node_prob_var(parent_node)
                or_list.append(parent_node_prob == 1)
            self.constraints.append(z3.Implies(node_prob == 0, z3.Or(or_list, self.context), self.context))


def open_interval_constraint(prob_var, context, low=0, high=1):
    return z3.And(prob_var > low, prob_var < high, context)


class LabCmp(Semantics):
    def generate_constraints(self):
        labAdm = LabAdm(self.z3i)
        self.constraints.extend(labAdm.constraints)
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            node_undec = open_interval_constraint(node_prob, self.context)
            or_list = []
            for parent_node in node.get_parents():
                parent_node_prob = self.z3i.get_node_prob_var(parent_node)
                self.constraints.append(z3.Implies(node_undec, parent_node_prob < 1, self.context))
                parent_undec = open_interval_constraint(parent_node_prob, self.context)
                or_list.append(parent_undec)
            self.constraints.append(z3.Implies(node_undec, z3.Or(or_list, self.context), self.context))


class SLabCmp(Semantics):
    def generate_constraints(self):
        sLabAdm = SLabAdm(self.z3i)
        self.constraints.extend(sLabAdm.constraints)
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.is_initial():
                continue
            node_undec = open_interval_constraint(node_prob, self.context)
            or_list = []
            for parent_node in node.get_parents():
                parent_node_prob = self.z3i.get_node_prob_var(parent_node)
                self.constraints.append(z3.Implies(node_undec, parent_node_prob < 1, self.context))
                parent_undec = open_interval_constraint(parent_node_prob, self.context)
                or_list.append(parent_undec)
            self.constraints.append(z3.Implies(node_undec, z3.Or(or_list, self.context), self.context))


class PrAdm(Semantics):
    """
    Probabilistic admissibility semantics: CF and for all arguments B with attackers A1...An and attackees C1...Cm,
    P(C1 or ... or Cm) <= P(A1 or ... or An) holds. Equivalently, by negating each side and simplifying, we get
    P(nC1, ..., nCm) >= P(nA1, ..., nAn).
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            node_parents = node.get_parents()
            if node_parents:
                attacker_assignment = [(attacker, False) for attacker in node_parents]
                attacker_prob_var = self.z3i.get_prob_var(attacker_assignment)
            else:
                # there are no attackers, so the prob of any attacker holding is 0
                # thus, in the negated view, the prob is 1
                attacker_prob_var = 1
            node_children = node.get_children()
            if node_children:
                attackee_assignment = [(attackee, False) for attackee in node_children]
                attackee_prob_var = self.z3i.get_prob_var(attackee_assignment)
            else:
                # there are no attackees, so the prob of any attackee holding is 0
                # thus, in the negated view, the prob is 1
                attackee_prob_var = 1
            if node_parents or node_children:
                self.constraints.append(attackee_prob_var >= attacker_prob_var)


class MinAdm(Semantics):
    """
    Min-admissibility semantics: CF and for every argument C, P(C) <= min_{B in Pre(C)} P(OR Pre(B)) holds.
    Equivalently, for all B in Pre(C) with Pre(B) = {A1, ..., An}, it holds P(C) <= 1 - P(nA1, ..., nAn).
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            node_parents = node.get_parents()
            if node_parents:
                node_prob = self.z3i.get_node_prob_var(node)
                for attacker in node_parents:
                    attacker_parents = attacker.get_parents()
                    if attacker_parents:
                        assignment = [(defender, False) for defender in attacker_parents]
                        prob_var = self.z3i.get_prob_var(assignment)
                        self.constraints.append(node_prob <= 1 - prob_var)
                    else:
                        self.constraints.append(node_prob == 0)
            else:
                pass  # minAdm enforces no constraint for initial arguments


class JntAdm(Semantics):
    """
    Joint-admissibility semantics: CF and for every argument C, P(C) <= P(AND_{B in Pre(C)} (OR Pre(B))) holds.
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) <= 1, so we skip it
                continue
            node_prob = self.z3i.get_node_prob_var(node)
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                self.constraints.append(node_prob == 0)
                continue

            prob_vars = self.z3i.positive_cnf_probvars(cnf)
            # self.constraints.append(node_prob <= z3.Sum(prob_vars))
            self.constraints.append(node_prob.__le__(z3.Sum(prob_vars)))


class ElmAdm(Semantics):
    """
    Element-wise admissibility semantics: If a probabilistic variable from the full joint distribution has positive
    probability, it needs to correspond to an admissible extension in the classical sense. That is, all prob vars
    corresponding to non-admissible extensions need to have probability zero.
    """
    def generate_constraints(self):
        sAdm = classical.Admissible(self.af)
        adm_extensions = sAdm.get_extensions()
        for assignment in self.z3i.full_joint_assignments:
            assignment_ext = set([node for (node, b) in assignment if b])
            if assignment_ext not in adm_extensions:
                prob_var = self.z3i.get_prob_var(assignment)
                self.constraints.append(prob_var == 0)


# Completeness

class WCmp(Semantics):
    """
    Weak completeness semantics: CF and p_A = 1 if and only if A is almost-surely defended.
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.is_initial():
                # every initial node is almost-surely defended
                self.constraints.append(node_prob == 1)
                continue
            as_defense_constraint = almost_sure_defense_constraint(self.z3i, node)
            self.constraints.append(z3.Implies(node_prob == 1, as_defense_constraint, ctx=self.context))
            self.constraints.append(z3.Implies(as_defense_constraint, node_prob == 1, ctx=self.context))


def prCmp_constraint(nodes: List[AF.Node], z3i: zi.Z3Instance) -> List[z3.BoolRef]:
    """
    Given a list of arguments, returns the constraints P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) for every argument C.
    """
    constraints = []
    for node in nodes:
        node_prob = z3i.get_node_prob_var(node)
        if node.is_initial():
            # for nodes without attackers, the constraint becomes P(C) >= 1
            constraints.append(node_prob == 1)
            continue
        # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
        cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
        if [] in cnf:
            # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
            # the constraint would be P(C) >= 0, so we skip it
            continue

        prob_vars = z3i.positive_cnf_probvars(cnf)
        constraints.append(node_prob >= z3.Sum(prob_vars))
    return constraints


class PrCmp(Semantics):
    """
    Probabilistic completeness semantics: Prob. admissible and P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) holds for all
    arguments C.
    """
    def generate_constraints(self):
        prAdm = PrAdm(self.z3i)
        self.constraints.extend(prAdm.constraints)
        self.constraints.extend(prCmp_constraint(self.af.get_nodes(), self.z3i))


class MinCmp(Semantics):
    """
    Min-complete semantics: Min-admissible and P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) holds for all arguments C.
    """
    def generate_constraints(self):
        minAdm = MinAdm(self.z3i)
        self.constraints.extend(minAdm.constraints)
        self.constraints.extend(prCmp_constraint(self.af.get_nodes(), self.z3i))


class JntCmp(Semantics):
    """
    Joint-complete semantics: CF and P(C) = P(AND_{B in Pre(C)} (OR Pre(B))) holds for all arguments C.
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) == 1
                self.constraints.append(node_prob == 1)
                continue
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                self.constraints.append(node_prob == 0)
                continue
            prob_vars = self.z3i.positive_cnf_probvars(cnf)
            # self.constraints.append(node_prob == z3.Sum(prob_vars))
            self.constraints.append(node_prob.__eq__(z3.Sum(prob_vars)))


class ElmCmp(Semantics):
    """
    Element-wise completeness semantics: If a probabilistic variable from the full joint distribution has positive
    probability, it needs to correspond to a complete extension in the classical sense. That is, all prob vars
    corresponding to non-complete extensions need to have probability zero.
    """
    def generate_constraints(self):
        sCmp = classical.Complete(self.af)
        cmp_extensions = sCmp.get_extensions()
        for assignment in self.z3i.full_joint_assignments:
            assignment_ext = set([node for (node, b) in assignment if b])
            if assignment_ext not in cmp_extensions:
                prob_var = self.z3i.get_prob_var(assignment)
                self.constraints.append(prob_var == 0)


#####################
# semantics by Käfer
#####################

class WNorS(Semantics):
    """
    Weak not-or semantics without CF: Generates constraints p_A <= p_nB1_nB2_..._nBi for every non-initial argument A
    with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            node_prob = self.z3i.get_node_prob_var(node)
            assignment = [(attacker, False) for attacker in node.get_parents()]
            attackers_prob = self.z3i.get_prob_var(assignment)
            # self.constraints.append(node_prob <= attackers_prob)
            self.constraints.append(node_prob.__le__(attackers_prob))


class NorS(Semantics):
    """
    Not-or semantics without CF: Generates constraints p_A = p_nB1_nB2_..._nBi for every non-initial argument A with
    attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            if node.is_initial():
                continue
            node_prob = self.z3i.get_node_prob_var(node)
            assignment = [(attacker, False) for attacker in node.get_parents()]
            attackers_prob = self.z3i.get_prob_var(assignment)
            # self.constraints.append(node_prob == attackers_prob)
            self.constraints.append(node_prob.__eq__(attackers_prob))


class SNorS(Semantics):
    """
    Strong not-or semantics without CF. Generates constraints p_A = 1 for initial arguments and p_A = p_nB1_nB2_..._nBi
    for every non-initial argument A with attackers B1, ..., Bi for some i>0.
    Also the intersection of Nor and Fou.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.is_initial():
                self.constraints.append(node_prob == 1)
            else:
                assignment = [(attacker, False) for attacker in node.get_parents()]
                attackers_prob = self.z3i.get_prob_var(assignment)
                # self.constraints.append(node_prob == attackers_prob)
                self.constraints.append(node_prob.__eq__(attackers_prob))


class WNor(Semantics):
    """
    Weak not-or semantics: Generates constraints for conflict-freeness and p_A <= p_nB1_nB2_..._nBi for every
    non-initial argument A with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            node_prob = self.z3i.get_node_prob_var(node)
            assignment = [(attacker, False) for attacker in node.get_parents()]
            attackers_prob = self.z3i.get_prob_var(assignment)
            # self.constraints.append(node_prob <= attackers_prob)
            self.constraints.append(node_prob.__le__(attackers_prob))


class Nor(Semantics):
    """
    Not-or semantics: Generates constraints for conflict-freeness and p_A = p_nB1_nB2_..._nBi for every non-initial
    argument A with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            if node.is_initial():
                continue
            node_prob = self.z3i.get_node_prob_var(node)
            assignment = [(attacker, False) for attacker in node.get_parents()]
            attackers_prob = self.z3i.get_prob_var(assignment)
            # self.constraints.append(node_prob == attackers_prob)
            self.constraints.append(node_prob.__eq__(attackers_prob))


class SNor(Semantics):
    """
    Strong not-or semantics: Generates constraints for conflict-freeness, p_A = 1 for initial arguments and
    p_A = p_nB1_nB2_..._nBi for every non-initial argument A with attackers B1, ..., Bi for some i>0.
    Also the intersection of Nor and Fou and CF.
    """
    def generate_constraints(self):
        sCF = CF(self.z3i)
        self.constraints.extend(sCF.constraints)
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.is_initial():
                self.constraints.append(node_prob == 1)
            else:
                assignment = [(attacker, False) for attacker in node.get_parents()]
                attackers_prob = self.z3i.get_prob_var(assignment)
                # self.constraints.append(node_prob == attackers_prob)
                self.constraints.append(node_prob.__eq__(attackers_prob))


class AF(Semantics):
    """
    Special semantics to add constraints for probability values or intervals specified in the AF input file.
    """
    def generate_constraints(self):
        for node in self.af.get_nodes():
            node_prob = self.z3i.get_node_prob_var(node)
            if node.value is not None:
                assert node.interval is None, "node with both value and interval set"
                # self.constraints.append(node_prob == node.value)
                self.constraints.append(node_prob.__eq__(node.value))
            elif node.interval is not None:
                low, high = node.interval
                # self.constraints.append(node_prob >= low)
                # self.constraints.append(node_prob <= high)
                self.constraints.append(node_prob.__ge__(low))
                self.constraints.append(node_prob.__le__(high))


class NNor(Semantics):
    """
    Noisy not-or semantics:  TODO
    per default without prior probabilities
    p_A_nB_nC = 1 * p_nB_nC
    p_A_nB_C  = (1 - t) * p_nB_C
    p_A_B_nC  = (1 - s) * p_B_nC
    p_A_B_C   = (1 - s ) * (1 - t) * p_B_C
    """
    def __init__(self, z3i: zi.Z3Instance, use_prior_probs=False):
        self.use_prior_probs = use_prior_probs
        super().__init__(z3i)

    def generate_constraints(self):
        # we use attack strength, so a prob_var should be created for each edge
        self.z3i.generate_edge_vars()

        for node in self.af.get_nodes():

            node_prior_prob_var = self.z3i.create_real_var("pr_" + node.name)  # e.g. pr_A
            if self.use_prior_probs:
                if node.value is not None:
                    # self.constraints.append(node_prior_prob_var == node.value)
                    self.constraints.append(node_prior_prob_var.__eq__(node.value))
                elif node.interval is not None:
                    i_min, i_max = node.interval
                    self.constraints.append(z3.And(node_prior_prob_var >= i_min, node_prior_prob_var <= i_max))
                else:
                    self.constraints.append(z3.And(node_prior_prob_var >= 0, node_prior_prob_var <= 1))
            else:
                self.constraints.append(node_prior_prob_var == 1)

            if node.is_initial():
                node_prob = self.z3i.get_node_prob_var(node)
                # self.constraints.append(node_prob == node_prior_prob_var)
                self.constraints.append(node_prob.__eq__(node_prior_prob_var))
                continue

            for parent_assignment in assignments.generate(node.get_parents()):
                parent_prob_var = self.z3i.get_prob_var(parent_assignment)  # e.g. p_nB_C
                prob_var = self.z3i.get_prob_var(parent_assignment + [(node, True)])  # e.g. p_A_nB_C

                product = node_prior_prob_var * parent_prob_var
                for (parent, status) in parent_assignment:
                    if status:
                        edge = node.get_parent_edge(parent)
                        edge_var = self.z3i.get_edge_var(edge)
                        product *= (1 - edge_var)
                # self.constraints.append(prob_var == product)
                self.constraints.append(prob_var.__eq__(product))


class NNorAF(NNor):
    """
    Noisy not-or semantics:  TODO
    with prior probabilities
    p_A_nB_nC = prA * p_nB_nC
    p_A_nB_C  = prA * (1 - t) * p_nB_C
    p_A_B_nC  = prA * (1 - s) * p_B_nC
    p_A_B_C   = prA * (1 - s ) * (1 - t) * p_B_C
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(z3i, use_prior_probs=True)


class CFs(Semantics):
    """
    Conflict-freeness semantics with attack strengths: p_A_B <= 1-s for all attacks A -s-> B.
    """
    def generate_constraints(self):
        # we use attack strength, so a prob_var should be created for each edge
        self.z3i.generate_edge_vars()

        for edge in self.af.get_edges():
            edge_var = self.z3i.get_edge_var(edge)
            assignment = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = self.z3i.get_prob_var(assignment)
            self.constraints.append(prob_var <= 1 - edge_var)


class StrengthCF(Semantics):
    """
    Alternative conflict-freeness semantics with attack strengths: "P(A,B) <= (1-s)P(A)" for all attacks A -s-> B.
    """
    def generate_constraints(self):
        # we use attack strength, so a prob_var should be created for each edge
        self.z3i.generate_edge_vars()

        for edge in self.af.get_edges():
            edge_var = self.z3i.get_edge_var(edge)
            assignment = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = self.z3i.get_prob_var(assignment)
            node_to_prob = self.z3i.get_node_prob_var(edge.node_to)
            self.constraints.append(prob_var <= (1 - edge_var) * node_to_prob)


########################################
# Strength and Support
########################################

class StrengthSupportCF(Semantics):
    """
    Treats arrows as support with strengths: "P(A,B) >= s * P(A)" for all supports A -s-> B.
    """
    def generate_constraints(self):
        # we use attack strength, so a prob_var should be created for each edge
        self.z3i.generate_edge_vars()

        for edge in self.af.get_edges():
            edge_var = self.z3i.get_edge_var(edge)
            assignment = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = self.z3i.get_prob_var(assignment)
            node_to_prob = self.z3i.get_node_prob_var(edge.node_to)
            # self.constraints.append(prob_var >= edge_var * node_to_prob)
            self.constraints.append(prob_var.__ge__(edge_var * node_to_prob))


########################################
# Dirac versions of classical semantics
########################################

class DiracClassicalSemantics(Semantics):
    """
    Lifts classical semantics into probabilistic semantics by admitting the Dirac-distributions corresponding to the
    classical semantics' extensions.
    E.g., a distribution P is Dirac-admissible if it is a Dirac distribution and P(beta) = 1 for some assignment beta
    implies that the set of arguments holding in beta is an admissible set.

    This is implemented by adding a constraint "P(beta_1) = 1 OR P(beta_2) = 1 OR ... OR P(beta_n) = 1" where beta_i are
    the assignments corresponding to the respective semantics' extensions. Note that this constraint also ensures that
    distributions satisfying it are Dirac distributions, so no additional Dirac constraints are added.
    """
    def __init__(self, classical_semantics: Type[classical.ClassicalSemantics], z3i: zi.Z3Instance):
        self.classical_semantics = classical_semantics
        super().__init__(z3i)

    def generate_constraints(self):
        cs = self.classical_semantics(self.af)
        or_list = []
        for extension in cs.get_extensions():
            assignment = [(node, node in extension) for node in self.af.get_nodes()]
            prob_var = self.z3i.get_prob_var(assignment)
            or_list.append(prob_var == 1)
        self.constraints.append(z3.Or(or_list))


class DiracCF(DiracClassicalSemantics):
    """
    Dirac-conflict-free semantics: Only allows Dirac-distributions corresponding to conflict-free extensions.
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(classical.CF, z3i)


class DiracAdm(DiracClassicalSemantics):
    """
    Dirac-admissible semantics: Only allows Dirac-distributions corresponding to admissible extensions.
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(classical.Admissible, z3i)


class DiracCmp(DiracClassicalSemantics):
    """
    Dirac-complete semantics: Only allows Dirac-distributions corresponding to complete extensions.
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(classical.Complete, z3i)


class DiracGrn(DiracClassicalSemantics):
    """
    Dirac-grounded semantics: Only allows Dirac-distributions corresponding to grounded extensions.
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(classical.Grounded, z3i)


class DiracPrf(DiracClassicalSemantics):
    """
    Dirac-preferred semantics: Only allows Dirac-distributions corresponding to preferred extensions.
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(classical.Preferred, z3i)


class DiracSStl(DiracClassicalSemantics):
    """
    Dirac-semi-stable semantics: Only allows Dirac-distributions corresponding to semi-stable extensions.
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(classical.SemiStable, z3i)


class DiracStl(DiracClassicalSemantics):
    """
    Dirac-stable semantics: Only allows Dirac-distributions corresponding to stable extensions.
    """
    def __init__(self, z3i: zi.Z3Instance):
        super().__init__(classical.Stable, z3i)


###################
# Helper functions
###################

def get_semantics_class_by_name(name: str):
    """
    By dark magic, get the semantics class given its name as string, e.g. "wNor" to yield the class WNor.

    :param name: the name of the semantics (case insensitive)
    :return: The semantics class corresponding to the name
    """
    queue = Semantics.__subclasses__().copy()
    while queue:
        semantics_class = queue.pop()
        if name.lower() == semantics_class.__name__.lower():
            return semantics_class
        queue.extend(semantics_class.__subclasses__())
    return None


def all_semantics_names():
    """
    :return: A list of the names of all semantics declared in this file
    """
    names = []
    # use a queue to recursively also get the names of subclasses
    queue = Semantics.__subclasses__().copy()
    while queue:
        semantics_class = queue.pop()
        queue.extend(semantics_class.__subclasses__())
        names.append(semantics_class.__name__)
    return names



from typing import NewType

import z3

import DPGM as AF
import z3_instance as zi

# label constants
Label = NewType("Label", str)
l_in = Label("in")
l_out = Label("out")
l_undec = Label("undec")


class LabelingScheme:

    num_args = 0  # The number of arguments each labeling scheme takes

    def get_label(self, argument: AF.Node, model: z3.ModelRef, z3i: zi.Z3Instance) -> Label:
        """
        Get the label of the given argument according the given model and the labeling approach.
        :return: l_in, l_out or l_undec
        """
        # implemented by subclasses
        pass

    def get_constraints(self, node: AF.Node, z3i: zi.Z3Instance):
        """
        Return a list of constraints on the probabilistic variable of the node where each constraint corresponds to a
        different label. I.e. for cautious labeling and node A, return [p_A == 1, p_A == 0, And(p_A < 1, p_A > 0)].
        Used to compute all labelings.
        """
        # implemented by subclasses
        return []


class Threshold(LabelingScheme):
    """
    Given two thresholds t_in and t_out, all nodes with probability
     - greater equal t_in are labeled 'in',
     - less than or equal to t_out are labeled 'out',
     - and between t_in and t_out are labeled 'undec'.
     """
    num_args = 2

    def __init__(self, t_in: float, t_out: float):
        self.t_in = t_in
        self.t_out = t_out

    def get_label(self, node: AF.Node, model, z3i: zi.Z3Instance):
        prob = z3i.get_node_value(node, model)
        if prob >= self.t_in:
            return l_in
        elif prob <= self.t_out:
            return l_out
        else:
            return l_undec

    def get_constraints(self, node: AF.Node, z3i: zi.Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob >= self.t_in
        out_constraint = node_prob <= self.t_out
        undec_constraint = z3.And(node_prob < self.t_in, node_prob > self.t_out, z3i.context)
        return [in_constraint, out_constraint, undec_constraint]


class Cautious(Threshold):
    """
    Arguments with probability 1 are labeled 'in', those with probability 0 are labeled 'out' and all others are
    labeled 'undec'.
    """
    num_args = 0

    def __init__(self):
        super().__init__(1.0, 0.0)


class Firm(LabelingScheme):
    """
    Arguments with probability 0.5 are labeled 'undec', those with probability >0.5 are labeled 'in' and all those with
    probability <0.5 are labeled 'out'.
    """
    def get_label(self, node: AF.Node, model, z3i: zi.Z3Instance):
        prob = z3i.get_node_value(node, model)
        if prob > 0.5:
            return l_in
        elif prob < 0.5:
            return l_out
        else:
            return l_undec

    def get_constraints(self, node: AF.Node, z3i: zi.Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob > 0.5
        out_constraint = node_prob < 0.5
        undec_constraint = node_prob == 0.5
        return [undec_constraint, in_constraint, out_constraint]
        # undec first because checking for equality is probably easier than inequalities


class ThresholdClassical(LabelingScheme):
    """
    Requires two thresholds t_in and t_min_attack_out. A node is labeled
     - 'in' if its probability is greater or equal t_in,
     - 'out' if it is not labeled 'in' and one of its attackers has probability greater or equal t_min_attack_out, and
     - 'undec' in all other cases.
     """
    num_args = 2

    def __init__(self, t_in: float, t_min_attack_out: float):
        self.t_in = t_in
        self.t_att_out = t_min_attack_out

    def get_label(self, node: AF.Node, model, z3i: zi.Z3Instance):
        node_prob = z3i.get_node_value(node, model)
        if node_prob >= self.t_in:
            return l_in
        else:
            label = l_undec
            for attacker in node.get_parents():
                attacker_prob = z3i.get_node_value(attacker, model)
                if attacker_prob >= self.t_att_out:
                    label = l_out
                    break
            return label

    def get_constraints(self, node: AF.Node, z3i: zi.Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        if self.t_in == self.t_att_out:
            in_constraint = node_prob >= self.t_in
            not_in_constraint = node_prob < self.t_in
            return [in_constraint, not_in_constraint]
        else:
            t_max = max(self.t_in, self.t_att_out)
            t_min = min(self.t_in, self.t_att_out)
            first_constraint = node_prob >= t_max
            second_constraint = node_prob < t_min
            third_constraint = z3.And(node_prob < t_max, node_prob >= t_min, z3i.context)
            return [first_constraint, second_constraint, third_constraint]


class Classical(ThresholdClassical):
    """
    A node is labeled
     - 'in' if it holds with probability 1,
     - 'out' if it is not labeled 'in' and one of its attackers holds with probability 1, and
     - 'undec' in all other cases.
    """
    num_args = 0

    def __init__(self):
        super().__init__(1.0, 1.0)


class Optimistic(LabelingScheme):
    """
    A node is labeled
     - 'in' if it holds with probability larger than 0, and
     - 'out' otherwise.
     """

    def get_label(self, node: AF.Node, model, z3i: zi.Z3Instance):
        prob = z3i.get_node_value(node, model)
        if prob > 0.0:
            return l_in
        else:
            return l_out

    def get_constraints(self, node: AF.Node, z3i: zi.Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob > 0
        out_constraint = node_prob == 0
        return [in_constraint, out_constraint]


class Pessimistic(LabelingScheme):
    """
    A node is labeled
     - 'in' if it holds with probability 1, and
     - 'out' otherwise.
     """

    def get_label(self, node: AF.Node, model, z3i: zi.Z3Instance):
        prob = z3i.get_node_value(node, model)
        if prob == 1.0:
            return l_in
        else:
            return l_out

    def get_constraints(self, node: AF.Node, z3i: zi.Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob == 1
        out_constraint = node_prob < 1
        return [in_constraint, out_constraint]


def get_labeling_scheme_class(name: str):
    """
    By dark magic, get the labeling scheme class given its name as string, e.g. "firm" to yield the class 'Firm'.

    :param name: the name of the labeling scheme class (case insensitive)
    :return: the labeling scheme class corresponding to the name, or None if no match was found
    """
    name = name.lower()
    queue = LabelingScheme.__subclasses__().copy()
    while queue:
        labeling_class = queue.pop()
        if name == labeling_class.__name__.lower():
            return labeling_class
        queue.extend(labeling_class.__subclasses__())
    return None


def get_all_labeling_scheme_names():
    """
    :return: a list of the names of all available labeling schemes
    """
    names = []
    queue = LabelingScheme.__subclasses__().copy()
    while queue:
        labeling_class = queue.pop()
        queue.extend(labeling_class.__subclasses__())
        names.append(labeling_class.__name__)
    return names



from typing import Dict, List, Tuple
# import os

import z3

import assignments
import DPGM as AF


class Z3Instance:
    """Class to bundle the declared probabilistic variables, the underlying full joint distribution."""

    def __init__(self, af: AF.DPGM, assume_independence=False):
        self.af = af
        self.assume_independence = assume_independence  # when this flag is True, no full joint probability is built

        self.full_joint_prob_vars: Dict[str, z3.ArithRef] = dict()
        self.full_joint_assignments: List[List[Tuple[AF.Node, bool]]] = []
        self.other_prob_vars: Dict[str, z3.ArithRef] = dict()
        self.real_vars: Dict[str, z3.ArithRef] = dict()
        self.constraints = set()

        self.edge_prob_vars: Dict[AF.Edge, z3.ArithRef] = dict()
        self.edge_constraints = set()
        self.edge_prob_vars_generated = False

        # create a z3 solver
        self.context = z3.Context()
        self.solver = z3.Solver(ctx=self.context)

        if not self.assume_independence:
            self._generate_full_prob_vars()
            # # serialize the constraints for the full joint distribution
            # path = os.path.splitext(self.af.path)[0]  # strip the .tgf extension
            # path1 = path + "_full-joint-probvars.py"
            # with open(path1, "w") as f:
            #     f.write(repr(self.full_joint_prob_vars))
            # print("Written to", path1)
            # path2 = path + "_constraints.py"
            # with open(path2, "w") as f:
            #     f.write(repr(self.constraints))

        # workaround to have the prob vars corresponding to the nodes' marginal probabilities in the models
        # returned by z3
        for node in self.af.get_nodes():
            self.get_node_prob_var(node)

    def generate_edge_vars(self):
        """
        Generates a probabilistic variable for each edge. If an edge has a value or interval given, those are added as
        constraints.
        """
        if self.edge_prob_vars_generated:
            # There might be several calls to generate_edge_vars, but we want to add the constraints only once.
            return
        for edge in self.af.get_edges():
            self.edge_prob_vars[edge] = z3.Real(edge.label, self.context)
            edge_prob_var = self.edge_prob_vars[edge]
            if edge.value is not None:
                self.edge_constraints.add(edge_prob_var == edge.value)
                assert edge.interval is None, "both edge.value and edge.interval where given"
            elif edge.interval is not None:
                (val_min, val_max) = edge.interval
                self.edge_constraints.add(z3.And(edge_prob_var >= val_min, edge_prob_var <= val_max, self.context))
            else:
                self.edge_constraints.add(edge_prob_var >= 0)
                self.edge_constraints.add(edge_prob_var <= 1)
        self.edge_prob_vars_generated = True

    def get_edge_var(self, edge: AF.Edge):
        if not self.edge_prob_vars_generated:
            self.generate_edge_vars()
        return self.edge_prob_vars[edge]

    def _generate_full_prob_vars(self):
        """Adds variables for the full joint distribution to the context. Adds the constraint that they sum to one."""
        constraint = 0
        if not self.full_joint_assignments:
            self.full_joint_assignments = assignments.generate(self.af.get_nodes())
        for assignment in self.full_joint_assignments:
            assignment_prob_name = prob_name(assignment)

            prob_var = z3.Real(assignment_prob_name, self.context)
            self.full_joint_prob_vars[assignment_prob_name] = prob_var
            self.constraints.add(prob_var >= 0)
            self.constraints.add(prob_var <= 1)
            constraint += prob_var

        constraint = 1 == constraint
        self.constraints.add(constraint)

    def get_node_prob_var(self, node: AF.Node):
        return self.get_prob_var([(node, True)])

    def get_prob_var(self, assignment: List[AF.Tuple[AF.Node, bool]]):
        """
        Get the probabilistic variable corresponding to a full or partial assignment.
        If it does not exist yet, it is created along with the constraints 'new_var >= 0', 'new_var <= 1' and the
        representation of new_var as sum of the full joint variables.
        """
        assert assignment, "empty assignment given"

        prob_var_name = prob_name(assignment)
        if prob_var_name in self.other_prob_vars:
            return self.other_prob_vars[prob_var_name]
        if prob_var_name in self.full_joint_prob_vars:
            return self.full_joint_prob_vars[prob_var_name]

        prob_var = z3.Real(prob_var_name, self.context)
        self.other_prob_vars[prob_var_name] = prob_var
        self.constraints.add(prob_var >= 0)
        self.constraints.add(prob_var <= 1)
        if not self.assume_independence:
            self.constraints.add(self.sum_out(assignment, prob_var))
        return prob_var

    def get_node_value(self, node: AF.Node, model: z3.ModelRef):
        """
        Get the value assigned to a node's probabilistic variable by the given model.
        :return: The value as float
        """
        return get_prob_var_value(self.get_node_prob_var(node), model)

    def create_real_var(self, name: str):
        """
        Creates and returns an auxiliary z3 real variable.
        """
        if name in self.real_vars:
            raise ValueError("A real var with name '" + name + "' already exists.")
        real_var = z3.Real(name, self.context)
        self.real_vars[name] = real_var
        return real_var

    def positive_cnf_probvars(self, clauses: List[List[AF.Node]]) -> List[z3.ArithRef]:
        """
        Given a positive CNF (i.e. without any negation in front of literals), return all prob vars of the full joint
        distribution which satisfy the CNF.
        E.g. for  (A OR B) AND (B OR C), given as [[A, B], [B, C]], return p_A_B_C, p_A_B_nC, p_nA_B_C, p_nA_B_nC and
        p_A_nB_C.
        :param clauses: A list of the conjunctive clauses of the positive CNF
        :return: A list of all prob vars from the full joint distribution which satisfy the CNF
        """
        candidates = [set()]
        given_arguments = set()
        for clause in clauses:
            given_arguments.update(clause)
            clause_assignments = assignments.generate(clause, [False, True])
            # Clauses contain disjunctive literals, so a true assignment is any assignment which assigns true at least
            # once. Filter out the first result from 'assignments' function because this assigns False to all literals.
            true_clause_assignments = clause_assignments[1:]

            new_candidates = []
            for candidate in candidates:
                # e.g. candidate = [(A, True), (B, False)]
                for true_clause_assignment in true_clause_assignments:
                    # e.g. true_clause_assignment = [(B, False), (C, True)]
                    # check whether the arguments appearing in both assignments are assigned the same value
                    skip = False
                    for (node, b) in true_clause_assignment:
                        if (node, not b) in candidate:
                            skip = True
                            break
                    # if so, add the new arguments to the candidate
                    if not skip:
                        new_candidate = candidate.copy()
                        new_candidate.update(true_clause_assignment)
                        new_candidates.append(new_candidate)
            candidates = new_candidates

        # arguments not present in the cnf can take on any value, so each previous candidate is extended by all possible
        # assignments of the remaining arguments
        remaining_arguments = set(self.af.get_nodes()).difference(given_arguments)
        remaining_assignments = assignments.generate(list(remaining_arguments))
        final_candidates = []
        for candidate in candidates:
            for remaining_assignment in remaining_assignments:
                final_candidate = candidate.copy()
                final_candidate.update(remaining_assignment)
                final_candidates.append(final_candidate)

        prob_vars = [self.get_prob_var(list(candidate)) for candidate in final_candidates]
        return prob_vars

    def sum_out(self, part_assignment: List[Tuple[AF.Node, bool]], part_assignment_prob_var):
        """[(A,True),(B,False)], [A,B,C] -> p_A_nB = p_A_nB_C + p_A_nB_nC"""
        part_assignment_args = [arg for (arg, _) in part_assignment]
        remaining_args = [arg for arg in self.af.get_nodes() if arg not in part_assignment_args]
        sum_list = []
        for assignment in assignments.generate(remaining_args):
            full_assignment = part_assignment + assignment
            full_assignment_name = prob_name(full_assignment)
            prob_var = self.full_joint_prob_vars[full_assignment_name]
            sum_list.append(prob_var)
        constraint = part_assignment_prob_var == z3.Sum(sum_list)
        return constraint
    
    def add_constraints(self, constraints):
        for cons in constraints:
            self.constraints.add(cons)

    def run(self, prob_constraints):
        """
        Add all constraints to the solver and check for a satisfying assignment.
        :param prob_constraints: a list of constraints
        :return: A z3 CheckSatResult, i.e. sat, unsat or unknown
        """
        self.solver.reset()
        self.solver.add(self.constraints)
        # print("self.constraints:", self.constraints)
        self.solver.add(self.edge_constraints)
        # print("self.af_constraints", self.af_constraints)
        self.solver.add(prob_constraints)
        # print("prob_constraints", prob_constraints)

        result = self.solver.check()
        # print(self.solver.to_smt2())
        # print(result)
        # if result == z3.sat:
        #     model = self.solver.model()
        #     self.print_distribution(model)
        #     # for node in self.af.get_nodes():
        #     #     prob = model.eval(self.get_node_prob_var(node))
        #     #     print(node.name, prob)
        return result

    def print_distribution(self, model: z3.ModelRef, nodes=None, only_positive=False):
        """
        Print distribution given by the model over the assignments of the given nodes.
        If no list of nodes is given, the distributions over all nodes in the AF is given.
        """
        if not nodes:
            nodes = self.af.get_nodes()
        node_assignments = assignments.generate(nodes)
        for assignment in node_assignments:
            prob_var = self.get_prob_var(assignment)
            value = get_prob_var_value(prob_var, model)
            if not only_positive or value > 0:
                assignments.print_assignment_prob(assignment, value)
                # print(prob_var, "=", value)

    def print_formatted_distribution(self, model: z3.ModelRef, format_string: str = ""):
        """
        Print distribution given by the model in the formats given by the format string: (M) marginal node
        probabilities, (F) full distribution, (S) support of distribution, (Z) Z3 output, (T) SMT2 format.
        """
        for char in format_string:
            if char == "M":
                for node in self.af.get_nodes():
                    print(node.name, "=", self.get_node_value(node, model))
            elif char == "F":
                self.print_distribution(model)
            elif char == "S":
                print("Support:")
                self.print_distribution(model, only_positive=True)
            elif char == "Z":
                print(model)
            elif char == "T":
                print(model.sexpr())
            else:
                print("Unknown distribution format '" + char + "'.")

    def generate_independence_constraints(self, a: List[AF.Node], b: List[AF.Node], c: List[AF.Node]):
        """
        Generate constraints stating that the nodes in A are independent of the nodes in B given the nodes in C.
        For each assignment a_A, a_B and a_C, we have
                 P(a_A | a_BC) = P(a_A | a_C)
        <=> P(a_ABC) / P(a_BC) = P(a_A | a_C)
        <=>           P(a_ABC) = P(a_BC) * P(a_A | a_C)   (this would suffice if P(a_A | a_C) is known from a CPT)
        <=>           P(a_ABC) = P(a_BC) * P(a_AC) / P(a_C)
        <=>  P(a_C) * P(a_ABC) = P(a_BC) * P(a_AC)
        """
        all_nodes = sorted(a + b + c)
        constraints = []
        for assignment in assignments.generate(all_nodes):
            a_ABC = assignment
            p_ABC = self.get_prob_var(a_ABC)
            a_BC = [(node, v) for (node, v) in assignment if node in b or node in c]
            p_BC = self.get_prob_var(a_BC)
            a_C = [(node, v) for (node, v) in assignment if node in c]
            if a_C:
                p_C = self.get_prob_var(a_C)
                a_AC = [(node, v) for (node, v) in assignment if node in a or node in c]
                p_AC = self.get_prob_var(a_AC)
                constraints.append(p_C * p_ABC == p_BC * p_AC)
            else:
                # a_C is empty, i.e., the constraint is an unconditional independence constraint
                a_A = [(node, v) for (node, v) in assignment if node in a]
                p_A = self.get_prob_var(a_A)
                constraints.append(p_ABC == p_BC * p_A)
        return constraints


def prob_name(assignment: List[AF.Tuple[AF.Node, bool]]) -> str:
    """
    Generates a fixed name for any assignment.
    Expects a list of tuples of nodes matched to True or False, i.e. [(node_B, False), (node_C, True)].
    Generates a name for the probability of the conjunction of the nodes, i.e. 'p_nB_C'.
    The order is determined by node IDs.
    """
    name = "p"
    assignment.sort()
    for (arg, b) in assignment:
        name += "_"
        if not b:
            name += "n"
        name += arg.name
    return name


def get_prob_var_value(prob_var, model: z3.ModelRef):
    """
    Get the value assigned to a probabilistic variable by the given model.
    :return: The value as float
    """
    # Based on https://stackoverflow.com/a/12600208/6620204
    value = model.eval(prob_var)
    if z3.is_int_value(value):
        return value.as_long()
    elif z3.is_rational_value(value):
        return float(value.numerator_as_long())/float(value.denominator_as_long())
    elif z3.is_algebraic_value(value):
        approx_value = value.approx(20)
        return float(approx_value.numerator_as_long())/float(approx_value.denominator_as_long())
    else:
        raise ValueError("Unable to convert to float: " + str(value))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cpraa']

package_data = \
{'': ['*'],
 'cpraa': ['.git/*',
           '.git/hooks/*',
           '.git/info/*',
           '.git/logs/*',
           '.git/logs/refs/heads/*',
           '.git/logs/refs/remotes/origin/*',
           '.git/objects/pack/*',
           '.git/refs/heads/*',
           '.git/refs/remotes/origin/*',
           '.idea/*',
           '.idea/inspectionProfiles/*',
           'AFs/*']}

install_requires = \
['z3-solver>=4.8.8,<5.0.0']

setup_kwargs = {
    'name': 'cpraa',
    'version': '0.2.1',
    'description': 'A checker for probabilistic abstract argumentation',
    'long_description': None,
    'author': 'Nikolai Käfer',
    'author_email': 'nikolai.kaefer@tu-dresden.de',
    'maintainer': 'Nikolai Käfer',
    'maintainer_email': 'nikolai.kaefer@tu-dresden.de',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import time


class Timer:

    all_timers = []
    enable_printing_default = True
    decimal_places_default = 3

    def __init__(self, start=True, output_string="Elapsed time:", enable_printing=None, decimal_places=None):
        self._start_time = None
        self._stop_time = None
        self.last_timing = None

        self.output_string = output_string

        if enable_printing is None:
            self.enable_printing = self.enable_printing_default
        else:
            self.enable_printing = enable_printing
        if decimal_places is None:
            decimal_places = self.decimal_places_default
        self.format_string = "{:." + str(decimal_places) + "f} seconds"

        self.all_timers.append(self)

        if start:
            self.start()

    def start(self):
        """Starts the timer"""
        self._start_time = time.perf_counter()

    def stop(self):
        """Stops the timer and returns the elapsed time in seconds"""
        self._stop_time = time.perf_counter()
        self.last_timing = self._stop_time - self._start_time
        if self.enable_printing:
            print(self.output_string, self.format_string.format(self.last_timing))
        return self.last_timing
